/* Metageek WiSPY Linux Tools
 *
 * LibUSB userspace control for the Metageek WiSPY spectrum analyzer
 * http://www.metageek.net/
 *
 * Writes a log file compatible with the MetaGeek WiSpy Windows tool
 *
 * Version 1.0
 * Mike Kershaw/Dragorn <dragorn@kismetwireless.net>
 * Joshua Wright <jwright@hasborg.com>
 *
 * This code is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Extra thanks to Ryan Woodings @ Metageek for interface documentation
 */

#include <stdio.h>
#include <usb.h>
#include <signal.h>
#include <time.h>
#include <string.h>

#include "config.h"

#ifdef HAVE_LINUX_DEVDISCONNECT
/* Needed for our own tools */
#include <unistd.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <linux/usbdevice_fs.h>
#include <errno.h>
#endif

/* First-gen firmware */
#define METAGEEK_WISPY1_VID	0x04b4
#define METAGEEK_WISPY1_PID	0x0bad

/* Second-gen firmware */
#define METAGEEK_WISPY2_VID	0x1781
#define METAGEEK_WISPY2_PID	0x083e

#define TIMEOUT 3000

/* Hid get report from spec */
#define HID_GET_REPORT 0x01
#define HID_RT_FEATURE 0x03

#define LOG_DESC "Description,"
#define LOG_HEADER "Wi-Spy Recording Created,"
#define LOG_COMMAS ",,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,"
#define LOG_CHAN "2400,2401,2402,2403,2404,2405,2406,2407,2408,2409,2410,2411,2412,2413,2414,2415,2416,2417,2418,2419,2420,2421,2422,2423,2424,2425,2426,2427,2428,2429,2430,2431,2432,2433,2434,2435,2436,2437,2438,2439,2440,2441,2442,2443,2444,2445,2446,2447,2448,2449,2450,2451,2452,2453,2454,2455,2456,2457,2458,2459,2460,2461,2462,2463,2464,2465,2466,2467,2468,2469,2470,2471,2472,2473,2474,2475,2476,2477,2478,2479,2480,2481,2482,2483,2484,"

/* Global for signal handling */
int sig = 0;

/* Basic device initialize, grabs either version 1 or version 2 */
static struct usb_device *init_wispy_device() {
	struct usb_bus *bus;
	struct usb_device *dev;

	/* Libusb init */
	usb_init();
	usb_find_busses();
	usb_find_devices();

	/* Scan all the devices on all the busses */
	for (bus = usb_busses; bus; bus = bus->next) {
		for (dev = bus->devices; dev; dev = dev->next) {
			if (((dev->descriptor.idVendor == METAGEEK_WISPY1_VID) &&
				 (dev->descriptor.idProduct == METAGEEK_WISPY1_PID)) ||
				((dev->descriptor.idVendor == METAGEEK_WISPY2_VID) &&
				 (dev->descriptor.idProduct == METAGEEK_WISPY2_PID)))
				return dev;
		}
	}

	return NULL;
}

#ifdef HAVE_LINUX_DEVDISCONNECT
/* Libusb doesn't seem to always provide this, so we'll use our own, taken from the 
* usb_detatch_kernel_driver_np...
*
* THIS IS A HORRIBLE EVIL HACK THAT SHOULDN'T BE DONE, EVER
* 
*/
struct local_usb_ioctl {
	int ifno;
	int ioctl_code;
	void *data;
};

struct ghetto_libusb_devhandle {
	int fd;
	/* Nooo... so bad. */
};

int usb_detach_kernel_driver(struct usb_dev_handle *dev, int interface) {
	struct local_usb_ioctl command;
	struct ghetto_libusb_devhandle *gdev;

	command.ifno = interface;
	command.ioctl_code = USBDEVFS_DISCONNECT;
	command.data = NULL;

	gdev = (struct ghetto_libusb_devhandle *) dev;

	if (ioctl(gdev->fd, USBDEVFS_IOCTL, &command) < 0) {
		fprintf(stderr, "Could not detach kernel driver from interface %d: %s",
			interface, strerror(errno));
		return -1;
	}

	return 0;
}
#endif

void sighand() {
	sig = 1;
}

void logspecmap(float *specmap, FILE *fp) {
	int i;
	for (i=0; i < 85; i++) {
		if (specmap[i] == 0) {
			/* Uninitialized value, print low signal */
			fprintf(fp, "-61,");
		} else {
			fprintf(fp, "%2.0f,", specmap[i]);
		}
	}
	/* this is bizarre, the CVS log file reports more frequencies than
	* the wispy device supports, this last value is always faked */
	fprintf(fp, "-61");
}

int main(int argc, char *argv[]) {
	struct usb_device *dev;
	struct usb_dev_handle *wispy;
	char buf[8], desctext[256], timestr[64];
	float specmap[86];
	int x, index;
	FILE *fp;
	struct tm *timefmt;
	time_t now;

	memset(buf, 0, sizeof(char) * 8);

	if (argc < 2) {
		printf("Usage: wispy_log <output file> [\"description\"]\n");
		return -1;
	}


	if (argc > 2) {
		strncpy(desctext, argv[2], sizeof(desctext)-1);
	} else {
		strncpy(desctext, "Created with wispy_log", 
				sizeof(desctext));
	}

	signal(SIGINT, sighand);
	signal(SIGTERM, sighand);
	signal(SIGQUIT, sighand);

	dev = init_wispy_device();
	if (!dev) {
		printf("Unable to find device.\n");
		return 1;
	}

	/* Open the wispy device */
	wispy = usb_open(dev);
	if (!wispy) {
		printf("Unable to open device\n");
		return 1;
	}

#if 1
#ifndef SYS_DARWIN
	/* Grab it like a keyboard interface, interface 0 */
	if (usb_claim_interface(wispy, 0) < 0) {
		printf("Unable to claim interface 0 -- are you root?\n"
			   "Is something else using it (another program or the kernel HID)?\n");
#ifdef HAVE_LINUX_DEVDISCONNECT
		printf("Using terrible hack to try to detach from the Linux HID...\n"
			   "This might not work at all\n");

		if (usb_detach_kernel_driver(wispy, 0) < 0) {
			printf("Failed to detatch from HID.\n");
			exit(1);
		}

		if (usb_claim_interface(wispy, 0) < 0) {
			printf("Seemed to detach from kernel but still couldn't claim interface.\n");
			exit(1);
		}

		printf("Detatched the USB interface from the kernel driver.\n");
#else
		exit(1);
#endif

	}
#endif
#endif

	if (*argv[1] == '-') {
		fp = stdout;
	} else {
		fp = fopen(argv[1], "w");
	}

	if (fp == NULL) {
		perror("fopen");
		return -1;
	}

	/* Create header records in file */
	fprintf(fp, "\n\n");
	time(&now);
	timefmt = localtime(&now);
	strftime(timestr, sizeof(timestr), "%c", timefmt);
	fprintf(fp, "%s%s%s\r\n", LOG_HEADER, timestr, LOG_COMMAS);
	fprintf(fp, "%s%s%s\r\n\r\n", LOG_DESC, desctext, LOG_COMMAS);
	fprintf(fp, "%s\r\n", LOG_CHAN);

	memset(&specmap, 0, sizeof(specmap));

	/* Loop and write signal statistics to fp */
	while (sig == 0) {


		/* Grab a report */
		if (usb_control_msg(wispy, USB_ENDPOINT_IN +
				USB_TYPE_CLASS + USB_RECIP_INTERFACE,
				HID_GET_REPORT, (HID_RT_FEATURE << 8),
				0, buf, 8, TIMEOUT) == 0) {
			printf("Failure in control_msg\n");
			return 1;
		}

		index = buf[0];
		for (x=1; x < 8; x++) {
			specmap[index+(x-1)] = ((float)buf[x] * 0.95f) - 61;
		}

		if (index == 77) {
			logspecmap(specmap, fp);
			fprintf(fp, "\r\n");
			memset(&specmap, 0, sizeof(specmap));
		}

	}

	/* Don't try to fclose stdout */
	if (*argv[1] != '-') {
		fclose(fp);
	}

	usb_close(wispy);

	return 0;
}	
